package payment

import (
	"context"
	"github.com/redis/go-redis/v9"
	"time"
)

type Payment struct {
	rdb *redis.Client
}

func NewPayment(redisConfig RedisConnParam) *Payment {
	return &Payment{
		rdb: redis.NewClient(&redis.Options{
			Addr:     redisConfig.Addr,
			Password: redisConfig.Password,
			DB:       redisConfig.DbIndex,
		}),
	}
}

func (p *Payment) IncrBalance(ctx context.Context, uid string, bal int64) (int64, error) {
	key := "balance:" + uid
	lockkey := "lk-bal" + uid
	for {
		locked, release, err := tryAcquire(ctx, p.rdb, lockkey, 5*time.Second)
		if err != nil {
			return 0, err
		}
		if !locked {
			continue
		}
		defer release()

		return p.rdb.IncrBy(ctx, key, bal).Result()
	}
}

func (p *Payment) GetBalance(ctx context.Context, uid string) (int64, error) {
	return p.rdb.Get(ctx, "balance:"+uid).Int64()
}

func (p *Payment) DecrBalance(ctx context.Context, uid string, bal int64) (int64, error) {
	key := "balance:" + uid
	lockkey := "lk-bal" + uid
	for {
		locked, release, err := tryAcquire(ctx, p.rdb, lockkey, 5*time.Second)
		if err != nil {
			return 0, err
		}
		if !locked {
			continue
		}
		defer release()

		return p.rdb.DecrBy(ctx, key, bal).Result()
	}
}

// also implement three method with key charge
func (p *Payment) IncrCharge(ctx context.Context, uid string, bal int64) (int64, error) {
	key := "charge:" + uid
	lockkey := "lk-char" + uid
	for {
		locked, release, err := tryAcquire(ctx, p.rdb, lockkey, 5*time.Second)
		if err != nil {
			return 0, err
		}
		if !locked {
			continue
		}
		defer release()

		return p.rdb.IncrBy(ctx, key, bal).Result()
	}
}

// implement GetCharge
func (p *Payment) GetCharge(ctx context.Context, uid string) (int64, error) {
	return p.rdb.Get(ctx, "charge:"+uid).Int64()
}

// implement DecrCharge
func (p *Payment) DecrCharge(ctx context.Context, uid string, bal int64) (int64, error) {
	key := "charge:" + uid
	lockkey := "lk-char" + uid
	for {
		locked, release, err := tryAcquire(ctx, p.rdb, lockkey, 5*time.Second)
		if err != nil {
			return 0, err
		}
		if !locked {
			continue
		}
		defer release()

		return p.rdb.DecrBy(ctx, key, bal).Result()
	}
}

// implement IncrCredits
func (p *Payment) IncrCredits(ctx context.Context, uid string, bal int64) (int64, error) {
	key := "credits:" + uid
	lockkey := "lk-credit" + uid
	for {
		locked, release, err := tryAcquire(ctx, p.rdb, lockkey, 5*time.Second)
		if err != nil {
			return 0, err
		}
		if !locked {
			continue
		}
		defer release()

		return p.rdb.IncrBy(ctx, key, bal).Result()
	}
}

// implement GetCredits
func (p *Payment) GetCredits(ctx context.Context, uid string) (int64, error) {
	return p.rdb.Get(ctx, "credits:"+uid).Int64()
}

// implement DecrCredits
func (p *Payment) DecrCredits(ctx context.Context, uid string, bal int64) (int64, error) {
	key := "credits:" + uid
	lockkey := "lk-credit" + uid
	for {
		locked, release, err := tryAcquire(ctx, p.rdb, lockkey, 5*time.Second)
		if err != nil {
			return 0, err
		}
		if !locked {
			continue
		}
		defer release()

		return p.rdb.DecrBy(ctx, key, bal).Result()
	}
}

func (p *Payment) Close() error {
	return p.rdb.Close()
}
